' rubiks cube
'
'	esc		quit
'	qawsed	turn width layers
'	rtfgvb	turn height layers
'	zuhjnm	turn depth layers
'	yxikop	turn whole cube
'	l			make a random move (for mixing the cube)
'	C			set cube back to solved state
'	1234567	set rotation speed to 2^n
'	.			detach/attach mouse
'	-			toggle center orientation markers

#Include "fbgfx.bi"
#Include "gl/gl.bi"
#Include "gl/glu.bi"

Const PI = 6.283185307178

Const SX = 512		' screen size
Const SY = 512

Const DD = 0.96	' small cube size (1 = no seams!)
Const FOV = 20		' vertical viewing angle
Const TR = 25		' distance, i.e. zoom, i.e. aparent size of cube
Const IND = 8		' initial rotation speed
Const MS = 500		' mouse sensitivity (pixels per pi/2)

Type vec
	As Double x, y, z
End Type

Type matrix
	As vec x, y, z
End Type

Type col
	As Double r, g, b
End Type

Type cube
	As vec ee(3)	' three unit vectors for small cube orientation
	As vec pp		' center of small cube
	As col co(6)	' color of all 6 small cube sides
End Type

Dim Shared As cube ru(3,3,3)		' whole cube = 3x3x3 small cubes
Dim Shared As Integer cs(4,2)		' side vertices in local coordinate system
Dim Shared As Integer ce(6,6)		' side coordinate system
Dim Shared As col scol(6)			' the 6 colors of the cube
Dim Shared As String sm				' move string
Dim Shared As Integer wcnt			' turn angle counter
Dim Shared As Double ND				' rotation speed
Dim Shared As Integer cent			' center orientation markers 0=off 1=on

Declare Sub main
main
End


Function col (r As Double, g As Double, b As Double) As col
	Dim As col c
	c.r = r
	c.g = g
	c.b = b
	Return c
End Function


Function vec (x As Double, y As Double, z As Double) As vec
	Dim As vec v
	v.x = x
	v.y = y
	v.z = z
	Return v
End Function


Function mal (c As Double, b As vec) As vec
	Dim As vec v
	v.x = c*b.x
	v.y = c*b.y
	v.z = c*b.z
	Return v
End Function


Function plus (a As vec, b As vec) As vec
	Dim As vec v
	v.x = a.x + b.x
	v.y = a.y + b.y
	v.z = a.z + b.z
	Return v
End Function


Function betrag (a As vec) As Double
	Return Sqr (a.x*a.x + a.y*a.y + a.z*a.z)
End Function


Function normiert (a As vec) As vec
	Dim As Double b
	b = betrag (a)
	If b=0 Then
		Return vec (0,0,0)
	Else
		Return mal (1/b, a)
	EndIf
End Function


Function skalar (a As vec, b As vec) As Double
	Return a.x*b.x + a.y*b.y + a.z*b.z
End Function


Function kreuz (a As vec, b As vec) As vec
	Dim As vec v
	v.x = a.y*b.z - a.z*b.y
	v.y = a.z*b.x - a.x*b.z
	v.z = a.x*b.y - a.y*b.x
	Return v
End Function


Function zwischenwinkel (a As vec, b As vec) As Double
	Return Acos (skalar (a, b) / (betrag (a)*betrag (b)))
End Function


Function mat_mal_vek (m As matrix, v As vec) As vec
	Dim As vec r
	r.x = skalar (m.x, v)
	r.y = skalar (m.y, v)
	r.z = skalar (m.z, v)
	Return r
End Function


Function zeilen_mat (a As vec, b As vec, c As vec) As matrix
	Dim As matrix r
	r.x = a
	r.y = b
	r.z = c
	Return r
End Function


Function spalten_mat (a As vec, b As vec, c As vec) As matrix
	Dim As matrix r
	r.x = vec (a.x, b.x, c.x)
	r.y = vec (a.y, b.y, c.y)
	r.z = vec (a.z, b.z, c.z)
	Return r
End Function


Function drehz_mat (w As Double) As matrix
	Dim As matrix r
	r.x = vec (Cos(w), -Sin(w), 0)
	r.y = vec (Sin(w),  Cos(w), 0)
	r.z = vec (0, 0, 1)
	Return r
End Function


' rotate vector a around vector b by the angle of w in radians
' return the new vector
'
Function drehen (a As vec, b As vec, w As Double) As vec
	Dim As vec nx, ny, nz, r
	Dim As Double z, phi, psi

	' establish a local coordinate system with b pointing in z direction
	' and two orthogonal unit vectors in x and y direction
	nz = normiert (b)
	Do
		phi = Rnd*PI
		psi = Rnd*PI/2-PI/4
		ny = vec (Cos(psi)*Cos(phi), Cos(psi)*Sin(phi), Sin(psi))
		z = zwischenwinkel (nz, ny)
	Loop While z<PI/4-PI/8 Or z>PI/4+PI/8
	nx = normiert (kreuz (ny, nz))
	ny = normiert (kreuz (nz, nx))

	' now we can simply make a coordinate transformation,
	' apply a rotation matrix in z, and transform back
	r = mat_mal_vek (zeilen_mat (nx, ny, nz), a)
	r = mat_mal_vek (drehz_mat (w), r)
	r = mat_mal_vek (spalten_mat (nx, ny, nz), r)

	Return r
End Function


Function outside (x As Integer, y As Integer, z As Integer, s As Integer) As Integer
	If x=2 And s=0 Or x=0 And s=3 Or z=2 And s=2 Or z=0 And s=5 Or y=0 And s=4 Or y=2 And s=1 Then Return 1
	Return 0
End Function


' draw all the small cubes, i.e. the whole cube
'
Sub drawall ()
	Dim As Integer x, y, z, s, v
	Dim As vec u, nx, ny, nz

	For x = 0 To 2
	For y = 0 To 2
	For z = 0 To 2
		For s = 0 To 5
			nx = mal (ce(s,0), ru(x,y,z).ee(ce(s,1)))
			ny = mal (ce(s,2), ru(x,y,z).ee(ce(s,3)))
			nz = mal (ce(s,4), ru(x,y,z).ee(ce(s,5)))
			
			glColor3f (ru(x,y,z).co(s).r, ru(x,y,z).co(s).g, ru(x,y,z).co(s).b)
			glBegin (GL_QUADS)
			For v = 0 To 3
				u = ru(x,y,z).pp
				u = plus (u, mal (DD/2, nx))
				u = plus (u, mal (cs(v,0)*DD/2, ny))
				u = plus (u, mal (cs(v,1)*DD/2, nz))
				glVertex3f (u.x, u.y, u.z)
			Next
			glEnd ()
			
			If cent Then
				If Abs(x-1)+Abs(y-1)+Abs(z-1)=1 And outside (x,y,z,s) Then
					glColor3f (0.8,0.8,0.8)
					glBegin (GL_QUADS)
					For v = 0 To 3
						u = ru(x,y,z).pp
						u = plus (u, mal (DD/2*1.01, nx))
						u = plus (u, mal (cs(v,0)*DD/2*0.3+0.3, ny))
						u = plus (u, mal (cs(v,1)*DD/2*0.1, nz))
						glVertex3f (u.x, u.y, u.z)
					Next
					glEnd ()
				EndIf
	
				If x=2 And y=2 And z=1 And s=0 Or x=1 And y=2 And z=2 And s=1 Or x=2 And y=1 And z=2 And s=2 Or x=0 And y=2 And z=1 And s=3 Or x=1 And y=0 And z=2 And s=4 Or x=2 And y=1 And z=0 And s=5 Then
					glColor3f (0.8,0.8,0.8)
					glBegin (GL_QUADS)
					For v = 0 To 3
						u = ru(x,y,z).pp
						u = plus (u, mal (DD/2*1.01, nx))
						u = plus (u, mal (cs(v,0)*DD/2*0.3-0.3, ny))
						u = plus (u, mal (cs(v,1)*DD/2*0.1, nz))
						glVertex3f (u.x, u.y, u.z)
					Next
					glEnd ()
				EndIf
			EndIf
			
		Next
	Next
	Next
	Next
End Sub


' initialize the small cubes to represent the solved state
'
Sub initru()
	Dim As Integer x, y, z, s

	For x = 0 To 2
	For y = 0 To 2
	For z = 0 To 2
		For s = 0 To 5
			If outside(x,y,z,s) Then
				ru(x,y,z).co(s) = scol(s)
			Else
				ru(x,y,z).co(s) = col(0,0,0)
			EndIf
		Next
		ru(x,y,z).ee(0) = vec(1,0,0)
		ru(x,y,z).ee(1) = vec(0,1,0)
		ru(x,y,z).ee(2) = vec(0,0,1)
		ru(x,y,z).pp = vec(x-1,y-1,z-1)
	Next
	Next
	Next
End Sub


' rotate everything in the given coordinate box around vector v by angle w
'
Sub drehru (xa As Integer, xe As Integer, ya As Integer, ye As Integer, za As Integer, ze As Integer, v As vec, w As Double)
	Dim As Integer x, y, z
	Dim As vec p
	
	For x = 0 To 2
	For y = 0 To 2
	For z = 0 To 2
		p = ru(x,y,z).pp
		If xa-0.5<p.x And p.x<xe+0.5 And ya-0.5<p.y And p.y<ye+0.5 And za-0.5<p.z And p.z<ze+0.5 Then
			ru(x,y,z).pp = drehen (ru(x,y,z).pp, v, w)
			ru(x,y,z).ee(0) = drehen (ru(x,y,z).ee(0), v, w)
			ru(x,y,z).ee(1) = drehen (ru(x,y,z).ee(1), v, w)
			ru(x,y,z).ee(2) = drehen (ru(x,y,z).ee(2), v, w)
		EndIf
	Next
	Next
	Next
End Sub


' take the first character from the move string and and do the move
' if the angle counter shows the move is done, delete this character
'
Sub keys ()
	Dim As Integer x, y, z
	Dim As Integer xa, xe, ya, ye, za, ze
	Dim As vec v
	Dim As Double w
	Dim As String i

	i = Mid(sm,1,1)


	If i="e" Then drehru (1, 1, -2, 2, -2, 2, vec(1,0,0), -PI/4/ND)
	If i="d" Then drehru (1, 1, -2, 2, -2, 2, vec(1,0,0),  PI/4/ND)

	If i="w" Then drehru (0, 0, -2, 2, -2, 2, vec(1,0,0), -PI/4/ND)
	If i="s" Then drehru (0, 0, -2, 2, -2, 2, vec(1,0,0),  PI/4/ND)

	If i="q" Then drehru (-1, -1, -2, 2, -2, 2, vec(1,0,0), -PI/4/ND)
	If i="a" Then drehru (-1, -1, -2, 2, -2, 2, vec(1,0,0),  PI/4/ND)


	If i="r" Then drehru (-2, 2, -2, 2, 1, 1, vec(0,0,1), -PI/4/ND)
	If i="t" Then drehru (-2, 2, -2, 2, 1, 1, vec(0,0,1),  PI/4/ND)

	If i="f" Then drehru (-2, 2, -2, 2, 0, 0, vec(0,0,1), -PI/4/ND)
	If i="g" Then drehru (-2, 2, -2, 2, 0, 0, vec(0,0,1),  PI/4/ND)

	If i="v" Then drehru (-2, 2, -2, 2, -1, -1, vec(0,0,1), -PI/4/ND)
	If i="b" Then drehru (-2, 2, -2, 2, -1, -1, vec(0,0,1),  PI/4/ND)

	
	If i="z" Then drehru (-2, 2, 1, 1, -2, 2, vec(0,1,0), -PI/4/ND)
	If i="u" Then drehru (-2, 2, 1, 1, -2, 2, vec(0,1,0),  PI/4/ND)

	If i="h" Then drehru (-2, 2, 0, 0, -2, 2, vec(0,1,0), -PI/4/ND)
	If i="j" Then drehru (-2, 2, 0, 0, -2, 2, vec(0,1,0),  PI/4/ND)

	If i="n" Then drehru (-2, 2, -1, -1, -2, 2, vec(0,1,0), -PI/4/ND)
	If i="m" Then drehru (-2, 2, -1, -1, -2, 2, vec(0,1,0),  PI/4/ND)
	
	
	If i="o" Then drehru (-2, 2, -2, 2, -2, 2, vec(0,0,1), -PI/4/ND)
	If i="p" Then drehru (-2, 2, -2, 2, -2, 2, vec(0,0,1),  PI/4/ND)

	If i="i" Then drehru (-2, 2, -2, 2, -2, 2, vec(1,0,0), -PI/4/ND)
	If i="k" Then drehru (-2, 2, -2, 2, -2, 2, vec(1,0,0),  PI/4/ND)

	If i="y" Then drehru (-2, 2, -2, 2, -2, 2, vec(0,1,0), -PI/4/ND)
	If i="x" Then drehru (-2, 2, -2, 2, -2, 2, vec(0,1,0),  PI/4/ND)
	
	
	If i<>"" Then
		wcnt += 1
		If wcnt=ND Then
			wcnt = 0
			sm = Mid(sm,2,Len(sm)-1)
		EndIf
	EndIf
End Sub


' is the character a valid move key (>0) or not (=0)
'
Function validkey (i As String) As Integer
	Return InStr("yxikopqawsedrtfgvbzuhjnm", i)
End Function


' generate a random move character, or sometimes two for a 180 turn
'
Function randommove () As String
	Dim As String s
	Dim As Integer t
	
	s = "qawsedrtfgvbzuhjnm"
	t = Int(Rnd*3*6)+1
	s = Mid(s,t,1)
	If Rnd<1/4 Then s = s+s

	Return s
End Function


' main
'
Sub main ()
	Dim As Double wx, wy
	Dim As Integer mx, my, buttons
	Dim As String i

	Randomize

	cs(0,0) = -1 : cs(0,1) = -1
	cs(1,0) =  1 : cs(1,1) = -1
	cs(2,0) =  1 : cs(2,1) =  1
	cs(3,0) = -1 : cs(3,1) =  1

	ce(0,0) = 1 : ce(0,1) = 0 : ce(0,2) = 1 : ce(0,3) = 1 : ce(0,4) = 1 : ce(0,5) = 2
	ce(1,0) = 1 : ce(1,1) = 1 : ce(1,2) = 1 : ce(1,3) = 2 : ce(1,4) = 1 : ce(1,5) = 0
	ce(2,0) = 1 : ce(2,1) = 2 : ce(2,2) = 1 : ce(2,3) = 0 : ce(2,4) = 1 : ce(2,5) = 1
	ce(3,0) =-1 : ce(3,1) = 0 : ce(3,2) = 1 : ce(3,3) = 1 : ce(3,4) =-1 : ce(3,5) = 2
	ce(4,0) =-1 : ce(4,1) = 1 : ce(4,2) = 1 : ce(4,3) = 2 : ce(4,4) =-1 : ce(4,5) = 0
	ce(5,0) =-1 : ce(5,1) = 2 : ce(5,2) = 1 : ce(5,3) = 0 : ce(5,4) =-1 : ce(5,5) = 1

	scol(0) = col(1,0,0)			' red
	scol(1) = col(0,0.8,0)		' green
	scol(2) = col(1,1,0)			' yellow
	scol(3) = col(1,0.6,0)		' orange
	scol(4) = col(0,0,1)			' blue
	scol(5) = col(1,1,1)			' white

	ScreenRes SX,SY,32,,FB.GFX_OPENGL+FB.GFX_MULTISAMPLE

	glViewport (0, 0, SX, SY)
	glMatrixMode (GL_PROJECTION)
	glLoadIdentity ()
	gluPerspective (FOV, SX/SY, TR-3, TR+3)
	glMatrixMode (GL_MODELVIEW)
	glLoadIdentity ()

	glClearColor (0.8, 0.6, 0.4, 1)		' background color
	glEnable (GL_DEPTH_TEST)
	glEnable (GL_CULL_FACE)
	'glPolygonMode (GL_FRONT_AND_BACK, GL_LINE)
	Dim As GLUquadric Ptr glq 
	glq = gluNewQuadric ()

	wx = 0		' angles for viewing direction
	wy = 0
	Sleep 10
	SetMouse SX/2, SY/2, 0, 1
	Sleep 10
	Flip
	Sleep 10
	GetMouse mx, my, , buttons

	ND = IND		' initial rotation speed
	cent = 0

	initru ()
	sm = ""
	wcnt = 0

	Do
		GetMouse mx, my, , buttons

		wx -= (mx-SX/2)/MS*PI/2
		wy -= (my-SY/2)/MS*PI/2
		SetMouse SX/2, SY/2, 0, 1
		If wx> PI/8 And wcnt=0 Then
			wx -= PI/4
			drehru (-2, 2, -2, 2, -2, 2, vec(0,0,1), -PI/4)
		EndIf
		If wx<-PI/8 And wcnt=0 Then
			wx += PI/4
			drehru (-2, 2, -2, 2, -2, 2, vec(0,0,1),  PI/4)
		EndIf
		If wy> PI/4 Then wy =  PI/4
		If wy<-PI/4 Then wy = -PI/4

		glClear (GL_COLOR_BUFFER_BIT Or GL_DEPTH_BUFFER_BIT)

		glPushMatrix ()
		glTranslatef (0, 0, -TR)
		glRotatef (-wy/PI*360-90, 1, 0, 0)
		glRotatef (-wx/PI*360, 0, 0, 1)
		
		drawall ()
		
		glColor3f (0,0,0)
		gluSphere (glq, 1+DD/2*0.99, 16, 16)
		
		glPopMatrix ()

		i = InKey$
		
		If i=Chr(27) Then Exit Do
		
		If i="." Then
			SetMouse SX/2, SY/2, 1 ,0
			Do : Flip : Loop Until InKey$="."
			SetMouse SX/2, SY/2, 0 ,1
		EndIf

		If i="-" Then cent = Not cent

		If i="C" Then
			initru ()
			sm = ""
			wcnt = 0
		EndIf

		If i>="1" And i<="7" And wcnt=0 Then
			ND = 2^(Asc(i)-Asc("0"))
		EndIf

		If validkey(i) Then sm = sm+i

		If i="l" Then sm = sm+randommove ()

		keys ()
		
		Flip
		
	Loop

	gluDeleteQuadric (glq)
	Screen 0

End Sub
